// @dart=2.12

import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:args/command_runner.dart';
import 'package:console/console.dart';
import 'package:git/git.dart';
import 'package:meta/meta.dart';

import '../models/app_version.dart';

// ignore: import_of_legacy_library_into_null_safe
import '../release.dart';

final runInShell = Platform.isWindows ? true : false;

/// The [T] is the type of the [parent].
abstract class BaseCommand<T> extends Command<T> {
  /// Get the [bumpType] from global results.
  ///
  /// See [kBumpTypeName].
  String get bumpType => globalResults![kBumpTypeName] as String;

  /// Get the [projectRootPath] from global results.
  ///
  /// See [kProjectRootName].
  String get projectRootPath => globalResults![kProjectRootName] as String;

  /// Get the project version of the project in [workingDirectoryPath].
  @protected
  Future<String> getProjectVersion(String workingDirectoryPath) {
    return Process.run(
      'cider',
      [
        'version',
        '--project-root',
        workingDirectoryPath,
      ],
      stdoutEncoding: utf8,
      runInShell: runInShell,
    ).then((value) => (value.stdout as String).trim());
  }

  @protected
  Future<GitDir> getGitDir(String workingDirectoryPath) async {
    if (await GitDir.isGitDir(workingDirectoryPath)) {
      return GitDir.fromExisting(workingDirectoryPath);
    } else {
      throw 'Invalid git project';
    }
  }

  @visibleForOverriding
  Future<void> formatFiles() async {
    print('Start formatting files...');
    final process = await Process.start(
      'flutter',
      'format lib'.split(' '),
    );
    final lineStream =
        process.stdout.transform(Utf8Decoder()).transform(LineSplitter());
    await for (final line in lineStream) {
      print(line);
    }

    await process.stderr.drain();
  }

  /// Bump the version of project in [projectPath].
  @protected
  Future<void> bumpVersionAsync({
    required String projectPath,
    bool bumpBuild = true,
  }) async {
    final ciderProjectRootArgName = '--project-root';
    await Process.run(
      'cider',
      'bump $bumpType -b $ciderProjectRootArgName $projectPath'.split(' '),
      stdoutEncoding: utf8,
      runInShell: runInShell,
    );
    if (bumpBuild) {
      await Process.run(
        'cider',
        'bump build $ciderProjectRootArgName $projectPath'.split(' '),
        stdoutEncoding: utf8,
        runInShell: runInShell,
      );
    }
  }

  @protected
  Future<void> createTagAsync({
    required GitDir gitDir,
    required String pubspecYamlPath,
  }) async {
    final currentBranch = await gitDir.currentBranch();
    if (![
      'master',
      'main',
    ].contains(currentBranch.branchName)) {
      throw 'Current branch is not master.';
    }

    print('Ensuring working directory clean...');
    if (!await gitDir.isWorkingTreeClean()) {
      throw 'Current branch is not clean.';
    }

    print('Pulling changes...');
    await gitDir.runCommand('pull'.split(' '));
    // Create git tag

    print('Start creating tag...');
    final version = AppVersion.parse(await getProjectVersion(pubspecYamlPath));

    final tagName = 'v${version.buildName}';
    final tagCreationOutput = Process.runSync(
      'git',
      'tag $tagName'.split(' '),
      stdoutEncoding: utf8,
      runInShell: runInShell,
    );
    if (tagCreationOutput.exitCode != 0) {
      throw tagCreationOutput.stderr;
    } else {
      print('Done creating tag...');

      const yesOption = 'yes';
      const noOption = 'no';

      final chooser = Chooser(
        [yesOption, noOption],
        message: 'Do you want to push this tag to remote?\n',
      );
      final response = await chooser.choose().then((value) => value as String);
      if (response == yesOption) {
        print('Start pushing tag to remote...');
        final pushTagToOriginResult = Process.runSync(
          'git',
          'push origin $tagName'.split(' '),
          stdoutEncoding: utf8,
          runInShell: runInShell,
        );
        if (pushTagToOriginResult.exitCode == 0) {
          print('Done pushing tag $tagName to remote...');
        } else {
          throw pushTagToOriginResult.stderr;
        }
      }
    }
  }
}
