// @dart=2.12

import 'dart:io';

import 'package:path/path.dart' as path;
// ignore: import_of_legacy_library_into_null_safe
import 'package:pubspec_yaml/pubspec_yaml.dart';

// ignore: import_of_legacy_library_into_null_safe
import '../release.dart';
import 'base_command.dart';

class PackageCommand extends BaseCommand {
  static const _platformInterfaceName = 'platform-interface';
  static const _kTagName = 'tag';

  @override
  String get description =>
      '''Specify to update version for a package that has platform interface.''';

  @override
  String get name => 'package';

  PackageCommand() {
    argParser.addOption(
      _platformInterfaceName,
      help: 'Specify the name of the platform interface name.',
    );

    argParser.addFlag(
      _kTagName,
      help: 'Create a tag in local.\n* Take note: You must in master branch.',
      defaultsTo: false,
      negatable: false,
    );
  }

  @override
  Future<void> run() async {
    if (argResults == null || globalResults == null) {
      throw 'Something when wrong when attempting to parse argument result.';
    }

    if (globalResults!.wasParsed(kBumpTypeName)) {
      await formatFiles();
      await _updateVersionAsync();
    } else if (argResults!.wasParsed(_kTagName)) {
      await createTagAsync(
        gitDir: await getGitDir(Directory.current.path),
        pubspecYamlPath: projectRootPath,
      );
    } else {
      throw ArgumentError(
          'You must at least specify $kBumpTypeName or $_kTagName.');
    }

    print('Done');
  }

  Future _updateVersionAsync() async {
    const pubspecFileName = 'pubspec.yaml';
    final projectRootDir = Directory(projectRootPath);
    final projectParentPath = projectRootDir.parent.path;

    // Hard code to use the parent as the GitDir. This is the best practice
    // for the project structure of an Endorsed Federated Plugin provided by
    // Flutter.

    final gitDir = await getGitDir(projectParentPath);

    print('Ensuring working directory clean...');
    if (!await gitDir.isWorkingTreeClean()) {
      throw 'Current branch is not clean.';
    }

    final projectRootPubspecFile = File(path.joinAll([
      projectRootPath,
      pubspecFileName,
    ]));

    final currentVersion = await getProjectVersion(projectRootPath);
    await bumpVersionAsync(
      projectPath: projectRootPath,
      bumpBuild: false,
    );
    final newVersion = await getProjectVersion(projectRootPath);

    print('Version updated from $currentVersion to $newVersion.');

    print('Start updating related section in current pubspec.yaml.');
    if (!argResults!.wasParsed(_platformInterfaceName)) {
      exit(0);
    }

    final platformInterfacePackageName =
        argResults![_platformInterfaceName] as String;

    var pubspec = PubspecYaml.loadFromYamlString(
        await projectRootPubspecFile.readAsString());
    final webPackageName = pubspec.customFields['flutter']!['plugin']
        ['platforms']['web']['default_package'] as String;

    final latestVersionTag = 'v$newVersion';
    for (final packageName in [webPackageName, platformInterfacePackageName]) {
      PubspecYaml _bumpPackageDependencyVersion({
        required PubspecYaml pubspec,
        required String packageName,
      }) {
        final dependencies = pubspec.dependencies.toList();
        final dependency =
            dependencies.firstWhere((x) => x.package() == packageName).iswitcho(
                  otherwise: () => null,
                  git: (gitPackageDependency) => gitPackageDependency.copyWith(
                    ref: gitPackageDependency.ref.map((_) => latestVersionTag),
                  ),
                )!;

        dependencies[
                dependencies.indexWhere((x) => x.package() == packageName)] =
            PackageDependencySpec.git(dependency);
        return pubspec.copyWith(dependencies: dependencies);
      }

      // Update the version in project root pubspec.

      pubspec = _bumpPackageDependencyVersion(
        pubspec: pubspec,
        packageName: packageName,
      );

      // Update the version in packages pubspec.

      final packageProjectPath = path.joinAll([projectParentPath, packageName]);
      await bumpVersionAsync(
        projectPath: packageProjectPath,
        bumpBuild: false,
      );

      // Bump the version of package platform interface package.

      if (packageName != platformInterfacePackageName) {
        final packageFile = File(path.joinAll([
          packageProjectPath,
          pubspecFileName,
        ]));
        var packagePubspec =
            PubspecYaml.loadFromYamlString(await packageFile.readAsString());
        packagePubspec = _bumpPackageDependencyVersion(
          pubspec: packagePubspec,
          packageName: platformInterfacePackageName,
        );

        await packageFile.writeAsString(packagePubspec.toYamlString());
      }
    }

    await projectRootPubspecFile.writeAsString(pubspec.toYamlString());
  }
}
